/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.ste.rdstmc.junit;

import static org.junit.Assert.assertTrue;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.ste.rdstmc.allgemein.Beschleunigung;
import de.bsvrz.ste.rdstmc.junit.tools.JUnitHelper;
import de.bsvrz.ste.rdstmc.junit.tools.ResultWrapper;
import de.bsvrz.ste.rdstmc.meldungserstellung.VerwaltungMeldungen;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Statische Definitionen zur Benutzung in den JUnit-Tests
 * 
 * @author Sans
 *
 */
public class RDSTMCPrueffall 
{
  /**
   * Debug-Logger fr Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();
  
  
  protected final static long KORREKTUR = 100;
  
  protected final static long TOL_PLUS = 15000;
  protected final static long TOL_MINUS = 5000;
  
  protected final static String ATG_PID = "atg.rdsMeldung";
  protected final static String ASP_PID = "asp.rdsGeneriert";
  
  
  private static long _startZeitPunkt = 0;
  
  protected static long getStartZeitpunkt()
  {
    if (_startZeitPunkt == 0)
    {
      throw new UnsupportedOperationException("Startzeitpunkt ist nicht initialisiert");
    }
    return _startZeitPunkt;
  }
  
  protected static void setStartZeitpunkt(long startZeitPunkt)
  {
    if (_startZeitPunkt > 0)
    {
      throw new UnsupportedOperationException("Startzeitpunkt ist bereits initialisiert");
    }
    if (startZeitPunkt <= 0)
    {
      throw new IllegalArgumentException("Startzeitpunkt mu grer als 0 sein");
    }
    _startZeitPunkt = startZeitPunkt;
  }
  
  protected static void setTestGenerator(Thread testGenerator)
  {
    _testGenerator = testGenerator;
  }
  
  private static Thread _testGenerator = null;
  
  protected static void assertTestGeneratorLaeuft()
  {
    assertTrue("Testgenerator muss laufen", _testGenerator != null && _testGenerator.isAlive());
  }
  
  protected static void assertPruefZeitKorrekt(long pruefZeitSoll, long pruefZeitIst)
  {
    long relPruefZeit = (pruefZeitIst - getStartZeitpunkt()) * Beschleunigung.WERT; 
    _debug.config("Zeitpunkt des Prfereignisses: " + relPruefZeit + "ms");
    assertTrue("Zeitpunkt des Prfereignisses zwischen " + (pruefZeitSoll - TOL_MINUS) + "ms" +
        " und " + (pruefZeitSoll + TOL_PLUS) + "ms",
        relPruefZeit >= (pruefZeitSoll - TOL_MINUS) && relPruefZeit <= (pruefZeitSoll + TOL_PLUS));
  }
  
  protected static void assertMeldungsDatenKorrekt(ResultData rd,
      int informationsTyp, 
      String meldungsArt,
      int geografischeBedeutung,
      int zeitlicheDringlichkeit,
      int zeitlicheDauer,
      int ausmassRichtung,
      int ereignisCode,
      int primaereLocation,
      String optionalKennung,
      String optionalWert)
  {
    ResultWrapper rw = new ResultWrapper(rd);
    
    assertTrue("InformationsTyp hat Wert " + informationsTyp +
        (informationsTyp == InformationsTyp.ALERTC ? " (fr Alert-C)" : ""),
        rw.getInformationsTyp() == informationsTyp);
    
    assertTrue("Meldungsart ist " + meldungsArt,
        rw.getMeldungsArt().equals(meldungsArt));
    
    assertTrue("Geografische Bedeutung hat Wert " + geografischeBedeutung +  
        (geografischeBedeutung == GeografischeBedeutung.NATIONAL ? " (fr national)" : ""),
        rw.getGeografischeBedeutung() == GeografischeBedeutung.NATIONAL);
    
    assertTrue("Zeitliche Dringlichkeit hat Wert " + zeitlicheDringlichkeit + 
        (zeitlicheDringlichkeit == ZeitlicheDringlichkeit.NORMAL ? " (fr normal)" : ""),
        rw.getZeitlicheDringlichkeit() == zeitlicheDringlichkeit);
    
    assertTrue("Zeitliche Dauer hat Wert " + zeitlicheDauer + 
        (zeitlicheDauer == ZeitlicheDauer.UNBEKANNT ? " (fr unbekannt)" : ""),
        rw.getZeitlicheDauer() == zeitlicheDauer);
    
    assertTrue("Ausma-/Richtung ist " + ausmassRichtung,
        rw.getAusmassRichtung() == ausmassRichtung);
    
    assertTrue("Ereigniscode ist " + ereignisCode,
        rw.getEreignisCode() == ereignisCode);
    
    assertTrue("Primaere TMC-Location ist " + primaereLocation,
        rw.getPrimaereLocation() == primaereLocation);
    
    if (optionalKennung != null)
    {
      assertTrue("Wert mit (optionaler) Kennung " + optionalKennung + " ist " + optionalWert,
          rw.getOptionalWert(optionalKennung).equals(optionalWert));
    }
  }
  
  protected static MutableSet getMeldungsMenge()
  {
    return JUnitHelper.getInstance().getLms().getMutableSet(VerwaltungMeldungen.MENGE);
  }
  
  protected static ClientDavInterface getConnection()
  {
    return JUnitHelper.getInstance().getConnection();
  }
  
  protected static DataModel getDataModel()
  {
    return getConnection().getDataModel();
  }
  
  private static DataDescription _dataDescription = 
    new DataDescription(getDataModel().getAttributeGroup(ATG_PID), getDataModel().getAspect(ASP_PID));
  
  public static DataDescription getDataDescription()
  {
    return _dataDescription;
  }
  
  public static class InformationsTyp
  {
    public final static int ALERTC = 0;
  }
  
  public static class MeldungsArt
  {
    public final static String AKTUALISIERUNG = "Modifikation";
    public final static String AUFHEBUNG = "Aufhebung";
  }
  
  public static class GeografischeBedeutung
  {
    public final static int NATIONAL = 3;
  }
  
  public static class ZeitlicheDringlichkeit
  {
    public final static int NORMAL = 0;
  }
  
  public static class ZeitlicheDauer
  {
    public final static int UNBEKANNT = 0;
  }
  
}
