/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.ste.rdstmc.situation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Verwaltung von Baustellen innerhalb eines Objekts vom Typ typ.netz.
 * 
 * @author Stefan Sans
 */
public class VerwaltungBaustellen extends AbstraktNetzVewDynObj<Situation>
  implements Runnable
{
  /**
   * Debug-Logger fr Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();
  
  public final static String MENGE = "Baustellen";
  
  public final static String MENGEN_TYP = "typ.verkehrsModellNetz";
    
  /**
   * Datenverteiler-Verbindung
   */
  private ClientDavInterface _connection;
  
  /**
   * @return die Datenverteiler-Verbindung
   */
  public ClientDavInterface getConnection()
  {
    return _connection;
  }
  
  private LinkedHashMap<Baustelle, BaustellenEigenschaften> _queue =
    new LinkedHashMap<Baustelle, BaustellenEigenschaften>();

  private void put(Baustelle baustelle, BaustellenEigenschaften daten)
  {
     synchronized (_queue)
     {
       _queue.remove(baustelle);
       _queue.put(baustelle, daten);
       _queue.notify();
     }
   } 
  
  private Map.Entry<Baustelle, BaustellenEigenschaften> take()
     throws InterruptedException
   {
    synchronized (_queue)
    {
      while (_queue.isEmpty())
      {
        _queue.wait();
      }
      Iterator<Map.Entry<Baustelle, BaustellenEigenschaften>> it= 
        _queue.entrySet().iterator();
      Map.Entry<Baustelle, BaustellenEigenschaften>  result = it.next();
      it.remove();
      return result;
    }
  }
  
  private Collection<SituationEigenschaftenListener> _listener = 
    new ArrayList<SituationEigenschaftenListener>();
  
   
  /**
   * Erzeugt eine neue Baustellen-Verwaltung
   * 
   * @param netz das Konfigurationsobjekt vom Typ typ.netz
   */
  public VerwaltungBaustellen(ClientDavInterface connection, ConfigurationObject netz,
      SituationEigenschaftenListener[] listener)
  {
    _connection = connection;
    
    if (listener != null)
    {
      for (SituationEigenschaftenListener stauEigenschaftenListener: listener)
      {
        _listener.add(stauEigenschaftenListener);
      }
    }  
    
    Thread thread = new Thread(this);
    thread.setName("Baustellen-Aktualisierungs-Thread");
    thread.start();
    _debug.config(thread + " gestartet");
    
    if (!initialisiereNetz(netz, MENGEN_TYP, MENGE))
    {
      _debug.warning("Das bergebene Netz " + netz.getName() +
      " enthlt KEINE Unternetze vom Typ " + MENGEN_TYP);
    }
  }
  
  @Override
  protected Situation erzeugeObjekt(SystemObject obj)
  {
    Baustelle baustelle = new Baustelle(getConnection(), obj);
    
    for (SituationEigenschaftenListener listener: _listener)
    {
      baustelle.addSituationsEigenschaftenListener(listener);
    }
    
    return baustelle; 
  }
  
  @Override
  protected Situation bereinigeObjekt(SystemObject obj)
  {
    new BaustellenBereinigungsThread(obj).start();
    return super.bereinigeObjekt(obj);
  }
     
  public void run()
  {
    while(true)
    {
      try
      {
        Map.Entry<Baustelle, BaustellenEigenschaften> entry = take();
        Baustelle baustelle = entry.getKey();
        BaustellenEigenschaften daten = entry.getValue();
        
        baustelle.fireSituationsEigenschaftenListener(daten);
      }
      catch (InterruptedException e)
      {
        _debug.warning("Thread wurde unerwartet unterbrochen", e);
      }
    }
  }
    
  class Baustelle extends Situation
  {
    /**
     * Datenbeschreibung fr Datensatz der Attributgruppe atg.situationsEigenschaften
     */
    protected final DataDescription _baustellenEigenschaftenBeschreibung;
    
      
    /**
     * Konstruiert eine Instanz vom Typ Baustelle und fhrt die Empfangsanmeldung
     * durch.
     * 
     * @param connection Datenverteiler-Verbindung
     * @param davObjekt Systemobjekt vom Typ Situation
     */
    public Baustelle(ClientDavInterface connection, SystemObject davObjekt)
    {
      super(connection, davObjekt);
      
      AttributeGroup atg = _connection.getDataModel().getAttributeGroup("atg.baustellenEigenschaften");
      Aspect asp = _connection.getDataModel().getAspect("asp.parameterSoll");
      _baustellenEigenschaftenBeschreibung = new DataDescription(atg, asp);
      
      getConnection().subscribeReceiver(this, getDavObjekt(), _baustellenEigenschaftenBeschreibung,
        ReceiveOptions.normal(), ReceiverRole.receiver());
      _debug.finer(_baustellenEigenschaftenBeschreibung + " zum Empfang angemeldet fr " + getDavObjekt());
    }
    
    /**
     * Gibt belegte Resourcen frei
     */
    void dispose()
    {
      _connection.unsubscribeReceiver(this, getDavObjekt(), _baustellenEigenschaftenBeschreibung);
      _debug.finer("Empfang abgemeldet von " + _baustellenEigenschaftenBeschreibung + " fr " + getDavObjekt());
    }
    
    /*
     * (Kein Javadoc)
     * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
     */
    public void update(ResultData[] resultData)
    {
      super.update(resultData);
      
      for (ResultData data : resultData)
      {
        Aspect asp = data.getDataDescription().getAspect();
        AttributeGroup atg = data.getDataDescription().getAttributeGroup();
        
        if (atg.equals(_baustellenEigenschaftenBeschreibung.getAttributeGroup()) &&
            asp.equals(_baustellenEigenschaftenBeschreibung.getAspect())) 
        {
           _debug.fine("Update Daten (Parameter) " + atg.getPid() + " fr " + this);
         
           uebernehmeDaten(data); ; 
        }
      }
    }
    
    @Override
    public void uebernehmeDaten(ResultData resultData)
    {
      AttributeGroup atg = resultData.getDataDescription().getAttributeGroup();
      if (atg.equals(_situationsEigenschaftenBeschreibung.getAttributeGroup()))
      {
        setSituationEigenschaften(resultData.hasData() ? resultData.getData() : null);
      }
      else if (atg.equals(_baustellenEigenschaftenBeschreibung.getAttributeGroup()))
      {
        setBaustellenEigenschaften(resultData.hasData() ? resultData.getData() : null); 
      }
      
      Data situationEigenschaften = getSituationsEigenschaften();
      Data baustellenEigenschaften = getBaustellenEigenschaften();
      
      BaustellenEigenschaften daten = null;
      
      if (situationEigenschaften != null &&  baustellenEigenschaften != null)
      {
        daten = new BaustellenEigenschaften(situationEigenschaften, baustellenEigenschaften);
      }
      
      put(this, daten);
    }
    
    /**
     * Aktueller Datensatz der Attributgruppe atg.situationsEigenschaften fr dieses Objekt
     */
    private Data _baustellenEigenschaften;
    
    /**
     * Zugriff auf die aktuellen Daten
     * 
     * @return den aktueller Datensatz der Attributgruppe atg.situationsEigenschaften
     * fr dieses Objekt
     */
    protected Data getBaustellenEigenschaften()
    {
      return _baustellenEigenschaften;
    }
    
    protected void setBaustellenEigenschaften(Data situationEigenschaften)
    {
      _baustellenEigenschaften = situationEigenschaften;
    }
  }
  
  class BaustellenBereinigungsThread extends Thread
  {
    BaustellenBereinigungsThread(SystemObject obj)
    {
      _obj = obj; 
    }
    
    private SystemObject _obj;
    
    public void run()
    {
      Situation baustelle = getVerwalteteObjekte().get(_obj.getPid());
      if (baustelle != null)
      {
        for (SituationEigenschaftenListener listener: _listener)
        {
          listener.neueSituationEigenschaften(baustelle, null);
          baustelle.removeSituationsEigenschaftenListener(listener);
        }
        baustelle.dispose();
      }
    }
  }
  
  static class BaustellenEigenschaften extends SituationEigenschaften
  {
    public final static int GUELTIG = 2;
    
    BaustellenEigenschaften(Data situationEigenschaften, Data baustellenEigenschaften)
    {
      super(situationEigenschaften);
      _status = baustellenEigenschaften.getItem("Status").asUnscaledValue().intValue();
    }
    
    private int _status = -1;
    
    @Override
    public boolean istGueltig(long aktuelleZeit)
    {
      return _status == GUELTIG && super.istGueltig(aktuelleZeit);
    }
    
  }
  
}
