/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.ste.rdstmc.test;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

public class StoerfallIndikator implements ClientSenderInterface
{
  /**
   * Debug-Logger fr Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();
  
  /**
   * Ist auf true gesetzt, solange auf Sendesteuerung gewartet wird
   */
  private boolean _warten = false;
  
  /**
   * Rolle, mit der als Sender angemeldet wird
   */
  public static final SenderRole ROLLE = SenderRole.source();
  
  /**
   * Zeitin ms, die max. auf Sendesteuerung gewartet wird
   */
  public static final long SENDE_TIMEOUT = 5000;
  
  /**
   * Ist diese Instanz zum DAV-Senden angemeldet?
   */
  private boolean _angemeldet = false;
  
  /**
   * @return ist diese Instanz zum DAV-Senden angemeldet?
   */
  public boolean isAngemeldet()
  {
    return _angemeldet;
  }
  
  public String toString()
  {
    return _davObjekt.getPid();
  }
  
  /**
   * Datenbeschreibung "generiert"
   */
  protected final DataDescription _datenBeschreibung;
  
  
  ClientDavInterface _connection;
  
  /**
   * @return die Datenverteiler-Verbindung
   */
  public ClientDavInterface getConnection()
  {
    return _connection;
  }
  
  private SystemObject _davObjekt;
  
  /**
   * @return das Dav-Systemobject, fr das das Senden von Daten erfolgt
   */
  public SystemObject getDavObjekt()
  {
    return _davObjekt;
  }
 
  /**
   * Konstruktor der Klasse
   * 
   * @param meldung Objekt vom Typ RDSVewMeldung, 
   * fr das das Senden von Daten erfolgt
   */
  public StoerfallIndikator(ClientDavInterface connection, SystemObject davObjekt,
      Aspect verfahren)
  {
    _connection = connection;
    _davObjekt = davObjekt;
       
    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.strfallZustand");
    _datenBeschreibung = new DataDescription(atg, verfahren);
    
    try
    {
      getConnection().subscribeSender(this, getDavObjekt(), _datenBeschreibung, ROLLE);
      _debug.fine(_datenBeschreibung + " zum Senden angemeldet fr " + getDavObjekt());
      _angemeldet = true;
    }
    catch (OneSubscriptionPerSendData e)
    {
      /* Es liegt bereits eine Sendeanmeldung von anderer Stelle vor
       * => hier nicht anmnelden
       */
      _debug.warning(getDavObjekt().getPid() + 
          " ist bereits an anderer Stelle zum Senden angemeldet", e);
    }
  }
  
  /**
   * Gibt die mit dieser Instanz verbundenen Ressourcen wieder frei
   */
  public void dispose()
  {
    if (_angemeldet)
    {
      getConnection().unsubscribeSender(this, getDavObjekt(), _datenBeschreibung);
      _debug.finer("Senden abgemeldet von " + _datenBeschreibung + " fr " + getDavObjekt());
      _angemeldet = false;
    }
  }
   
  /**
   * Unmittelbares Senden von Daten.
   * 
   * @param resultData die zu sendenden Daten
   * @return true, wenn das Senden erfolgreich war, false sonst
   */
  protected boolean send(ResultData resultData)
  {
    try
    {
      getConnection().sendData(resultData);
    }
    catch(SendSubscriptionNotConfirmed e)
    {
      if (SENDE_TIMEOUT > 0)
      {  
        _debug.finer("Warte max. " + SENDE_TIMEOUT + "ms auf Sendesteuerung fr " + getDavObjekt().getPid());
        synchronized(this)
        {
          _warten = true;
          try
          {
            wait(SENDE_TIMEOUT);
          }
          catch(InterruptedException ex)
          {
            // tue nichts
          }
          _warten = false;
        }
        try
        {
          getConnection().sendData(resultData);
        }
        catch (SendSubscriptionNotConfirmed ex)
        {
          _debug.warning("Keine Sendesteuerung fr " + getDavObjekt().getPid(), ex);
          return false;
        }
        return true;
      }
      else
      {
        _debug.warning("Keine Sendesteuerung fr " + getDavObjekt().getPid(), e);
        return false;
      }
    }
    catch (DataNotSubscribedException e)
    {
      _debug.error("Keine Sendeanmeldung bzw. Sendeanmeldung ungltig fr " + getDavObjekt().getPid(), e);
      return false;
    }
    return true;
  }
  
  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientSenderInterface#dataRequest(de.bsvrz.dav.daf.main.config.SystemObject, de.bsvrz.dav.daf.main.DataDescription, byte)
   */
  public void dataRequest(SystemObject object, DataDescription dataDescription, byte state)
  {
    _debug.finer("dataRequest() fr " + getDavObjekt().getPid() + " - Sender state: " + state);
    if (_warten && state == ClientSenderInterface.START_SENDING)
    {
      synchronized(this)
      {
        notify();
      }
    }
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientSenderInterface#isRequestSupported(de.bsvrz.dav.daf.main.config.SystemObject, de.bsvrz.dav.daf.main.DataDescription)
   */
  public boolean isRequestSupported(SystemObject so, DataDescription dd)
  {
    return true;
  }
  
  public void send(int zustandId)
  {
    this.
    send(zustandId, 1.0f);
  }
  
  public void send(int zustandId, float guete)
  {
    Data datensatz = _connection.createData(_datenBeschreibung.getAttributeGroup());
    datensatz.getItem("T").asTimeValue().setMillis(0);
    datensatz.getItem("Situation").asUnscaledValue().set(zustandId);
    datensatz.getItem("Horizont").asTimeValue().setMillis(0);
    Data data = datensatz.getItem("Gte");
    data.getItem("Index").asScaledValue().set(guete);
    data.getItem("Verfahren").asTextValue().setText("Standard");
    
    ResultData resultData = new ResultData(getDavObjekt(), _datenBeschreibung, 
        getConnection().getTime(), datensatz);
    send(resultData);
  }
  
}
