/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.ste.rdstmc.verkehrsanalyse;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ste.rdstmc.konfiguration.AeusseresStrassenSegment;
import de.bsvrz.ste.rdstmc.konfiguration.Konfiguration;
import de.bsvrz.ste.rdstmc.konfiguration.MessQuerschnitt;
import de.bsvrz.ste.rdstmc.konfiguration.StrassenSegment;
import de.bsvrz.ste.rdstmc.konfiguration.StrassenTeilSegment;
import de.bsvrz.ste.rdstmc.meldungserstellung.Route;
import de.bsvrz.ste.rdstmc.parameter.RDSParameterDaten;
import de.bsvrz.ste.rdstmc.parameter.RDSParameterListener;
import de.bsvrz.ste.rdstmc.parameter.RDSParameterVerwaltung;
import de.bsvrz.ste.rdstmc.situation.Situation;
import de.bsvrz.ste.rdstmc.situation.SituationEigenschaften;
import de.bsvrz.ste.rdstmc.situation.SituationEigenschaftenListener;
import de.bsvrz.ste.rdstmc.situation.VerwaltungStaus;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Instanziiert und verwaltet alle Strfallindikatoren
 * 
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.3 $ / $Date: 2009/01/27 13:54:22 $ / ($Author: Sans $)
 */
public class Verkehrsanalyse implements RDSParameterListener, 
  SituationEigenschaftenListener
{
  /**
   * Debug-Logger fr Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();
  
  /**
   * Mutex fr den Stoerfall-Thread und damit zu synchronisierende Threads
   */
  public final Object mutexStoerfall = new Object();
  
  /**
   * Mutex fr die Stoerfall-Situationen-Queue und damit zu synchronisierende Threads
   */
  public final Object mutexStoerfallSituation = new Object();;

  /**
   * Die DAV-Verbindung
   */
  private ClientDavInterface _connection;
  
  /**
   * @return die DAV-Verbindung
   */
  public ClientDavInterface getConnection()
  {
    return _connection;
  }
  
  /**
   * @return das Konfigurationsobjekt vom Typ typ.netz
   */
  public ConfigurationObject getNetz()
  {
    return getKonfiguration().getNetz();
  }
  
  /**
   *  Die Verwaltung der Konfigurationsobjekte des Netztes
   */
  private Konfiguration _konfiguration;
  
  public Konfiguration getKonfiguration()
  {
    return _konfiguration;
  }
  
  /**
   * Mapping der PID eines ueren Straensegments auf seine zugehrige
   * StoerfallSituation 
   */
  private Map<String, StoerfallSituation> _stoerfallSituationen = 
    new HashMap<String, StoerfallSituation> ();

  public StoerfallSituation getStoerfallSituation(String segmentPid)
  {
    return _stoerfallSituationen.get(segmentPid);
  }
  
  private RDSParameterDaten _parameter;
  
  public RDSParameterDaten getParameter()
  {
    return _parameter;
  }
  
  private void bestimmeStoerfallSituationen(Collection<Route> routen, boolean alleSegmente)
  {
    for (Route route: routen)
    {
      for (StrassenSegment segment: route.getSegmente())
      {
        if (alleSegmente || segment instanceof AeusseresStrassenSegment)
        {  
          StoerfallSituation stoerfallSituation =  _stoerfallSituationen.get(segment.getDavObjekt().getPid());
          if (stoerfallSituation == null)
          {  
            stoerfallSituation = new StoerfallSituation(segment, this);
            _stoerfallSituationen.put(segment.getDavObjekt().getPid(), stoerfallSituation);
            stoerfallSituation.addStoerfallSituationListener(route);
          }  
        }  
      }
    }
    _debug.config("Zur Verkehrsanalyse in " + getNetz().getPid() + " werden " + 
        _stoerfallSituationen.size() + " Strassensegmente benutzt");
  }
 
  private void bestimmeStoerfaelle(Collection<Route> routen)
  {
    int anzahlMqs = 0;
    int anzahlAbschnitte = 0;
    int anzahlSts = 0;
    
    for (StoerfallSituation stoerfallSituation: getStoerfallSituationen())
    {
      for (MessQuerschnitt mq: stoerfallSituation.getSegment().getMessQuerschnitte())
      {
        Stoerfall stoerfall = new LokalerStoerfall(mq, stoerfallSituation);
        stoerfallSituation.addStoerfall(stoerfall);
        stoerfall.addStoerfallListener(stoerfallSituation);
        anzahlMqs += 1;
      }
    }
    _debug.config("Zur Verkehrsanalyse in " + getNetz().getPid() + " werden " + 
        anzahlMqs + " Messquerschnitte benutzt");
        
    for (StoerfallSituation stoerfallSituation: getStoerfallSituationen())
    {
      StrassenSegment segment = stoerfallSituation.getSegment();
      for (StrassenTeilSegment sts: segment.getStrassenTeilSegmente())
      {
        Stoerfall stoerfall = new StreckenbezogenerStoerfall(sts, stoerfallSituation);
        stoerfallSituation.addStoerfall(stoerfall);
        stoerfall.addStoerfallListener(stoerfallSituation);
        anzahlSts += 1;
      }
    }
    _debug.config("Zur Verkehrsanalyse in " + getNetz().getPid() + " werden " + 
        anzahlSts + " Straenteilsegmente benutzt");
    
//    for (StrassenAbschnitt abschnitt: getKonfiguration().getStrassenAbschnitte().values())
//    {
//      StrassenSegment startSegment = abschnitt.getVonMq().getSegment();
//      StrassenSegment endSegment = abschnitt.getNachMq().getSegment();
//      
//      if (_stoerfallSituationen.get(startSegment.getDavObjekt().getPid()) != null &&
//          _stoerfallSituationen.get(endSegment.getDavObjekt().getPid()) != null)
//      {
//        // Finde Routen-Teiverlufe, die Start-und Endsegment in der richtigenReihenfolge enthalten
//        for (Route route: routen)
//        {
//          int startIndex = route.getSegmente().indexOf(startSegment);
//          int endIndex = route.getSegmente().indexOf(endSegment);
//          if (startIndex >= 0 && endIndex >= startIndex)
//          {
//            ListIterator<StrassenSegment> it = route.getSegmente().listIterator(startIndex);
//            while (it.hasNext() && startIndex <= endIndex)
//            {
//              StrassenSegment segment = it.next();
//              startIndex += 1;
//              StoerfallSituation situation = _stoerfallSituationen.get(segment.getDavObjekt().getPid());
//              if (situation != null)
//              {  
//                Stoerfall stoerfall = new StreckenbezogenerStoerfall(abschnitt, situation);
//                situation.addStoerfall(stoerfall);
//                stoerfall.addStoerfallListener(situation);
//                anzahlAbschnitte += 1;
//              }  
//            }
//          }
//        }
//      }
//      else if (_stoerfallSituationen.get(startSegment.getDavObjekt().getPid()) != null)
//      {   
//        // Finde Routen-Teiverlufe vom Startsegment bis Routenende
//        for (Route route: routen)
//        {
//          int startIndex = route.getSegmente().indexOf(startSegment);
//          if (startIndex >= 0)
//          {
//            ListIterator<StrassenSegment> it = route.getSegmente().listIterator(startIndex);
//            while (it.hasNext())
//            {
//              StrassenSegment segment = it.next();
//              StoerfallSituation situation = _stoerfallSituationen.get(segment.getDavObjekt().getPid());
//              if (situation != null)
//              {  
//                Stoerfall stoerfall = new StreckenbezogenerStoerfall(abschnitt, situation);
//                situation.addStoerfall(stoerfall);
//                stoerfall.addStoerfallListener(situation);
//                anzahlAbschnitte += 1;
//              }  
//            }
//          }
//        }
//      }
//      else if (_stoerfallSituationen.get(endSegment.getDavObjekt().getPid()) != null)
//      {   
//        // Finde Routen-Teiverlufe vom Routenende bis Endsegment
//        for (Route route: routen)
//        {
//          int startIndex = 0;
//          int endIndex = route.getSegmente().indexOf(endSegment);
//          if (endIndex >= 0)
//          {
//            Iterator<StrassenSegment> it = route.getSegmente().iterator();
//            while (it.hasNext() && startIndex <= endIndex)
//            {
//              StrassenSegment segment = it.next();
//              startIndex += 1;
//              StoerfallSituation situation = _stoerfallSituationen.get(segment.getDavObjekt().getPid());
//              if (situation != null)
//              {  
//                Stoerfall stoerfall = new StreckenbezogenerStoerfall(abschnitt, situation);
//                situation.addStoerfall(stoerfall);
//                stoerfall.addStoerfallListener(situation);
//                anzahlAbschnitte += 1;
//              }  
//            }
//          }
//        }
//      }
//    }
    _debug.config("Zur Verkehrsanalyse in " + getNetz().getPid() + " werden " + 
        anzahlAbschnitte + " Strassenabschnitte benutzt");
  }
      
  private VerwaltungStaus _stauVerwaltung;
  
  public Verkehrsanalyse(ClientDavInterface connection, Konfiguration konfiguration, 
      RDSParameterVerwaltung parameterVerwaltung, Collection<Route> routen, boolean alleSegmente)
  {
    super();
    
    _connection = connection;
    _konfiguration = konfiguration;
       
    parameterVerwaltung.addParameterListener(this);
    neueParameter(parameterVerwaltung, parameterVerwaltung.getAktuelleDaten());
    
    bestimmeStoerfallSituationen(routen, alleSegmente);
    _stauVerwaltung = new VerwaltungStaus(getConnection(), getNetz(), 
        parameterVerwaltung, new SituationEigenschaftenListener[] {this});
    
    bestimmeStoerfaelle(routen);
  }
  
  public Collection<StoerfallSituation> getStoerfallSituationen()
  {
    return Collections.unmodifiableCollection(_stoerfallSituationen.values());
  }
  
  public void neueParameter(RDSParameterVerwaltung quelle, RDSParameterDaten daten)
  {
    byte level = daten.getLevelFlags(getParameter());

    synchronized(mutexStoerfall)
    {
      synchronized(mutexStoerfallSituation)
      {
        _parameter = daten;
        
        if ((level & RDSParameterDaten.FLAG_LEVEL2) != 0)
        {
          _debug.fine("Neue Parameter: notify alle Strfall-Situationen");
          for (StoerfallSituation stoerfallSituation: getStoerfallSituationen())
          {
            stoerfallSituation.fireStoerfallSituationListener();
          }
        }
      }  
      if ((level & RDSParameterDaten.FLAG_LEVEL3) != 0)
      {
        _debug.fine("Neue Parameter: notify alle Strflle");
        for (StoerfallSituation stoerfallSituation: getStoerfallSituationen())
        {
          for (Stoerfall stoerfall: stoerfallSituation.getStoerFaelle())
          {  
            stoerfall.fireStoerfallListener();
          }  
        }
      }
    }
  }
  
  public void neueSituationEigenschaften(Situation quelle, SituationEigenschaften daten)
  {
    _debug.fine("Neue Staudaten fr " + quelle + ": " + (daten == null ? "<KEINE>" : "start = " + daten.startOffset + ", ende = " + daten.endOffset + 
        ", "+ daten.strassenSegment.length + " Segmente"));
    //Menge der zu benachichtigenden Segmente
    Set<StrassenSegment> segmente = new LinkedHashSet<StrassenSegment>();

    synchronized(mutexStoerfallSituation)
    {
      
//      for (StoerfallSituation stoerfallSituation: getStoerfallSituationen())
//      {
//        List<SegmentSituationDaten> liste = _segmentStauDaten.get(stoerfallSituation.getSegment())
//        
//        for (SegmentSituationDaten segmentDaten: getStauDaten())
//        {  
//          if (segmentDaten.situation == quelle)
//          {
//            alteSegmentDaten.add(segmentDaten);
//          }
//        }  
//      }
      
      
      // Entferne zunchst alle vorhanden Daten fr diese Situation
      Iterator<StrassenSegment> outer = _segmentStauDaten.keySet().iterator();
      while (outer.hasNext())
      {
        StrassenSegment segment = outer.next();
        List<SegmentSituationDaten> liste = _segmentStauDaten.get(segment);
        
        Iterator<SegmentSituationDaten> inner = liste.iterator();
        while (inner.hasNext())
        {
          SegmentSituationDaten segmentDaten = inner.next();
          if (segmentDaten.situation == quelle)
          {
            inner.remove();
            segmente.add(segment);
          }
        }
        if (liste.isEmpty())
        {
          outer.remove();
        }
      }
      
//      for (List<SegmentSituationDaten> liste: _segmentStauDaten.values())
//      {
//        Iterator<SegmentSituationDaten> it = liste.listIterator();
//        while (it.hasNext())
//        {
//          SegmentSituationDaten segmentDaten = it.next();
//          if (segmentDaten.situation == quelle)
//          {
//            it.remove();
//            segmente.add(segmentDaten.segment);
//          }
//        }
//      }
      // _segmentStauDaten.clear();
      
      // Fge neue Daten hinzu
      if (daten != null && daten.istGueltig(getConnection().getTime()))
      {  
        for (int i = 0; i < daten.strassenSegment.length; i += 1)
        {
          SystemObject objekt = daten.strassenSegment[i];
          StrassenSegment segment = getKonfiguration().getAeussereStrassenSegmente().get(objekt.getPid());
          if (segment == null)
          {
            segment = getKonfiguration().getInnereStrassenSegmente().get(objekt.getPid());
          }
          if (segment != null)
          {
            long start = i == 0 ? daten.startOffset : 0;
            long ende = i == daten.strassenSegment.length - 1 ? daten.endOffset : segment.getLaenge();
            
            List<SegmentSituationDaten> liste = _segmentStauDaten.get(segment);
            if (liste == null)
            {
              liste = new LinkedList<SegmentSituationDaten>();
              _segmentStauDaten.put(segment, liste);
            }
            liste.add(new SegmentSituationDaten(quelle, segment, start, ende));
            segmente.add(segment);
          }
        }
      }
      
      // Benachrichtige die betroffenen Segmente
      for (StrassenSegment segment: segmente)
      {
        StoerfallSituation stoerfallSituation = _stoerfallSituationen.get(segment.getDavObjekt().getPid());
        if (stoerfallSituation != null)
        {
          stoerfallSituation.fireStoerfallSituationListener();
        }
      }
    }    
  }
  
  
  /*
   * (Kein Javadoc)
   * @see de.bsvrz.ste.rdstmc.situation.SituationsEigenschaftenlListener
   * #neueSituationsEigenschaften(de.bsvrz.ste.rdstmc.situation.Stau) 
   */
//  public void neueSituationEigenschaftenAlt(Situation quelle, SituationEigenschaften daten)
//  {
//    synchronized(mutexStoerfallSituation)
//    {
//      _debug.fine("Neue Staudaten, start = " + daten.startOffset + ",ende = " + daten.endOffset + ", "+ daten.strassenSegment.length + "Segmente");
//      // Baue Menge der zu lschenden Daten
//      Set<SegmentSituationDaten> alteSegmentDaten = new LinkedHashSet<SegmentSituationDaten>();
//      for (StoerfallSituation stoerfallSituation: getStoerfallSituationen())
//      {
//        for (SegmentSituationDaten segmentDaten: getStauDaten(stoerfallSituation.getSegment()))
//        {  
//          if (segmentDaten.situation == quelle)
//          {
//            alteSegmentDaten.add(segmentDaten);
//          }
//        }  
//      }
//      
//      // Baue Menge der neu hinzugekommenen Daten
//      Set<SegmentSituationDaten> neueSegmentDaten = new LinkedHashSet<SegmentSituationDaten>();
//      // Nur Stauanteile positiver Lnge werden bercksichtigt
//      if (daten != null && daten.startOffset >= 0 && daten.endOffset > daten.startOffset)
//      {
//        for (SystemObject obj: daten.strassenSegment)
//        {  
//          StrassenSegment segment = null;
//          if (obj.isOfType("typ.ueresStraenSegment"))
//          {
//            segment = getKonfiguration().getAeussereStrassenSegmente().get(obj.getPid());
//          }
//          else if (obj.isOfType("typ.inneresStraenSegment"))
//          {
//            segment = getKonfiguration().getInnereStrassenSegmente().get(obj.getPid());
//          }
//            
//          if (segment != null)
//          {
//            long startOffset = 0;
//            if (obj == daten.strassenSegment[0])
//            {
//              startOffset = daten.startOffset <= segment.getLaenge()? daten.startOffset : segment.getLaenge();
//            }
//            long endOffset = segment.getLaenge();
//            if (obj == daten.strassenSegment[daten.strassenSegment.length - 1])
//            {
//              endOffset = daten.endOffset <= segment.getLaenge() ?  daten.endOffset : segment.getLaenge();
//            }
//            neueSegmentDaten.add(new SegmentSituationDaten(quelle, segment, startOffset, endOffset));
//          }
//        }  
//      }
//
//      // Optimiere Menge der zu lschenden Daten und Menge der neuern Daten mit Mengenarithmetik
//      Set<SegmentSituationDaten> segmentDaten =
//        new HashSet<SegmentSituationDaten>(alteSegmentDaten); 
//      segmentDaten.retainAll(neueSegmentDaten);
//      alteSegmentDaten.removeAll(segmentDaten);
//      neueSegmentDaten.removeAll(segmentDaten);
//    
//      // Menge der Pids der zustzlich zu benachrichtigenden Segmente
//      Set<String> segmente = new LinkedHashSet<String>();
//      String text = null;
//      
//      // Entferne zu lschenende Daten
//      for (SegmentSituationDaten alteDaten: alteSegmentDaten)
//      {
//        String key = alteDaten.segment.getDavObjekt().getPid();
//        StoerfallSituation stoerfallSituation = _stoerfallSituationen.get(key);
//        if (stoerfallSituation != null)
//        {
//          if (removeStauDaten(stoerfallSituation.getSegment(), alteDaten))
//          {
//            if (text == null)
//            {
//              text = "Entfernen Stauanteile von Stau " + alteDaten.situation;
//            }
//            text += "\n\tvon " + alteDaten.startOffset + "m bis " + alteDaten.endOffset +
//              "m in Segment " + key;
//            segmente.add(key);
//          }
//        }
//      }
//      
//      if (text != null)
//      {
//        _debug.fine(text);
//        text = null;
//      }
//      
//      // Fge neue Daten hinzu
//      for (SegmentSituationDaten neueDaten: neueSegmentDaten)
//      {
//        String key = neueDaten.segment.getDavObjekt().getPid();
//        StoerfallSituation stoerfallSituation = _stoerfallSituationen.get(key);
//        if (stoerfallSituation != null)
//        {
//          addStauDaten(stoerfallSituation.getSegment(), neueDaten);
//          if (text == null)
//          {
//            text = "Hinzufgen Stauanteile von Stau " + neueDaten.situation;
//          }
//          text += "\n\tvon " + neueDaten.startOffset + "m bis " + neueDaten.endOffset +
//            "m in Segment " + key;
//          segmente.add(key);
//        }
//      }
//      
//      if (text != null)
//      {
//        _debug.fine(text);
//        text = null;
//      }
//    
//      // Benachrichtige zustzlich die betroffenen Segmente
//      for (String segmentKey: segmente)
//      {
//        StoerfallSituation stoerfallSituation = _stoerfallSituationen.get(segmentKey);
//        if (stoerfallSituation != null)
//        {
//          stoerfallSituation.fireStoerfallSituationListener();
//        }
//      }
//    }
//  }
   
  private Map<StrassenSegment, List<SegmentSituationDaten>> _segmentStauDaten =
    new HashMap<StrassenSegment, List<SegmentSituationDaten>>();
  
  Collection<SegmentSituationDaten> getStauDaten(StrassenSegment segment)
  {
    Collection<SegmentSituationDaten> col = _segmentStauDaten.get(segment);
    if (col== null)
    {
      return Collections.emptySet();
    }
    else
    {
      return Collections.unmodifiableCollection(col);
    }
  }
  
  private boolean addStauDaten(StrassenSegment segment, SegmentSituationDaten daten)
  {
    List<SegmentSituationDaten> liste = _segmentStauDaten.get(segment);
    if (liste == null)
    {
      liste = new LinkedList<SegmentSituationDaten>();
      _segmentStauDaten.put(segment, liste);
    }
    return liste.add(daten);
  }
 
  private boolean removeStauDaten(StrassenSegment segment, SegmentSituationDaten daten)
  {
    Collection<SegmentSituationDaten> col = _segmentStauDaten.get(segment);
    if (col == null)
    {
      return false;
    }
    boolean result = col.remove(daten);
    if (col.isEmpty())
    {
      _segmentStauDaten.remove(segment);
    }
    return result;
  }
  
  
  public boolean existiertStau(StrassenSegment startSegment, long startOffset, StrassenSegment endSegment, 
      long endOffset)
  {
    StrassenSegment segment = startSegment;
    
    while (segment != null)
    {
      // Setze Start/-Ende-Offsets
      long start = segment == startSegment ? startOffset : 0;
      long ende = segment == endSegment ? endOffset : segment.getLaenge();
     
      synchronized (mutexStoerfallSituation)
      {
        Collection<SegmentSituationDaten> stauListe = _segmentStauDaten.get(segment);
        
        if (stauListe != null)
        {  
          // Iteriere Stau-Anteile fr das aktuelle Segment, Stau existiert bei berlappung
          for (SegmentSituationDaten stauDaten: stauListe)
          {
            if ((start >= stauDaten.startOffset && start < stauDaten.endOffset) ||
                (ende > stauDaten.startOffset && ende <= stauDaten.endOffset) ||
                (start < stauDaten.startOffset && ende > stauDaten.endOffset))
            {
              return true;
            }
          }
        }  
      }  
        
      // Hole nchstes Strassensegment wenn mglich
      if (segment == endSegment)
      {
        segment = null;
      }
      else
      {
        segment = segment.getKuerzestenNachfolgerAufStrasse();
      }
    }
    return false;
  }

  
  public boolean existiertStau(long startOffset, long endOffset, Collection<StrassenSegment> segmente)
  {
    boolean gefunden = false;
    
    LinkedList<StrassenSegment> verwalteteSegmente = new LinkedList<StrassenSegment>();
    
    for (StrassenSegment segment: segmente)
    {
      if (_stoerfallSituationen.get(segment.getDavObjekt().getPid()) != null)
      {
        verwalteteSegmente.add(segment);
      }
    }
    
    synchronized (mutexStoerfallSituation)
    {
      for (StrassenSegment segment: verwalteteSegmente)
      {
        for (SegmentSituationDaten stau: getStauDaten(segment))
        {
          long start = 0;
          long ende = segment.getLaenge();
          start = segment == verwalteteSegmente.getFirst() ? startOffset : start;
          ende = segment == verwalteteSegmente.getLast() ? endOffset : ende;
          
          if ((stau.startOffset >= start && stau.startOffset < ende) ||
              (stau.endOffset > start && stau.endOffset <= ende) ||
              (stau.startOffset < start && stau.endOffset > ende))
          {
            gefunden = true;
            break;
          }
        }
      }
    }
    
    return gefunden;
  }
}
